/**
 * $Id: RunBackgroundAspect.aj,v 1.2 2012/07/13 19:52:18 oboehm Exp $
 *
 * Copyright (c) 2008 by Oliver Boehm
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express orimplied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * (c)reated 19.01.2009 by oliver (ob@oasd.de)
 */
package patterntesting.concurrent;

import org.aspectj.lang.annotation.SuppressAjWarnings;
import org.slf4j.*;

import patterntesting.annotation.concurrent.RunBackground;
import patterntesting.runtime.util.*;

/**
 * This aspect together with the @RunBackground annotation allows you to
 * run a method in the backbround (as separate thread).
 *
 * @author <a href="boehm@javatux.de">oliver</a>
 * @since 0.9
 */
public aspect RunBackgroundAspect {

    private static final Logger log = LoggerFactory.getLogger(RunBackgroundAspect.class);

    pointcut backgroundMethods() :
        execution(@RunBackground * *..*.*(..))
        || execution(@RunBackground *..*.new(..));

    /**
     * Runs a method in the background.
     * 
     * @return always null
     */
    @SuppressAjWarnings({"adviceDidNotMatch"})
    Object around() : backgroundMethods() {
        if (!Environment.areThreadsAllowed()) {
            if (log.isDebugEnabled()) {
                log.debug(JoinPointHelper.getAsShortString(thisJoinPoint)
                        + " is NOT started in background (threading disabled)");                
            }
            return proceed();
        }
        Runnable work = new Runnable() {
            public void run() {
                proceed();
            }
        };
        if (log.isTraceEnabled()) {
            log.trace("starting "
                    + JoinPointHelper.getAsShortString(thisJoinPoint)
                    + " in background...");
        }
        try {
            runBackground(work);
        } catch (IllegalStateException ise) {
            log.warn("can't start "
                    + JoinPointHelper.getAsShortString(thisJoinPoint)
                    + " in background"
                    + " (reason: inside JEE server normally it is not allowed to start a thread)",
                    ise);
        } catch (RuntimeException re) {
            log.warn("can't start "
                    + JoinPointHelper.getAsShortString(thisJoinPoint)
                    + " in background", re);
            throw re;
        }
        return null;
    }

    private void runBackground(Runnable work) {
        Thread t = new Thread(work, "background");
        t.start();
    }

}
